//-------------------------------------------------------------------------------------
//              GHI Electronics, LLC
//               Copyright (c) 2008
//               All rights reserved
//-------------------------------------------------------------------------------------
/*
 * You can use this file if you agree to the following:
 *
 * 1. This header can't be changed under any condition.
 *    
 * 2. This is a free software and therefore is provided with NO warranty.
 * 
 * 3. Feel free to modify the code but we ask you to provide us with
 *	  any bugs reports so we can keep the code up to date.
 *
 * 4. This code may ONLY be used with GHI Electronics, LLC products.
 *
 * THIS SOFTWARE IS PROVIDED BY GHI ELECTRONICS, LLC ``AS IS'' AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED 
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL 
 * GHI ELECTRONICS, LLC BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR ORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  
 *
 *	Specs are subject to change without any notice
 */

/*
 * This example makes Embedded Master an HID USB Mouse. You can connect it to your PC and see the mouse cursor moves on the screen.
 * 
 */

using System;
using System.Threading;
using Microsoft.SPOT;
using Microsoft.SPOT.Hardware;
using Microsoft.SPOT.Hardware.UsbClient;

using Microsoft.SPOT.Input;
using Microsoft.SPOT.Presentation;
using Microsoft.SPOT.Presentation.Media;
using Microsoft.SPOT.Presentation.Controls;
using Microsoft.SPOT.Presentation.Shapes;

namespace USBdeviceTeset
{
    public class Program
    {
        static Bitmap screen = new Bitmap(SystemMetrics.ScreenWidth, SystemMetrics.ScreenHeight);
        static Font smallFont = Resources.GetFont(Resources.FontResources.small);

        public static void Main()
        {
            UsbStream mouseStream = null;
            UsbController[] controllers = null;
            try
            {
                screen.Clear();
                print("Initializaing EM as a USB", 0, 0);
                print("mouse.", 0, 10);
                screen.Flush();
                Thread.Sleep(2000); // make comment visible to user on screen

                // We should have one controller
                controllers = UsbController.GetControllers();

                // Device Descriptor, Vendor ID and Product ID are random numbers for this example
                Configuration.DeviceDescriptor deviceDesc = new Configuration.DeviceDescriptor(0x1234, 0x5678, 0x100);
                deviceDesc.bMaxPacketSize0 = 64;

                // Endpoint descriptors for a mouse
                Configuration.Endpoint[] epDesc = {
                new Configuration.Endpoint(1, Configuration.Endpoint.ATTRIB_Write | Configuration.Endpoint.ATTRIB_Interrupt),
                    };
                epDesc[0].bInterval = 10;   // poll time
                epDesc[0].wMaxPacketSize = 8;   // max packet size

                // Interface descriptors
                Configuration.UsbInterface[] interfaceDesc = {
                new Configuration.UsbInterface(0, epDesc),
                    };
                interfaceDesc[0].bInterfaceClass = 0x03;    // HID device
                interfaceDesc[0].bInterfaceSubClass = 0;
                interfaceDesc[0].bInterfaceProtocol = 0;

                // HID Class Descriptor
                interfaceDesc[0].classDescriptor = new Configuration.ClassDescriptor(0x21, hidClassDescriptorPayload);

                // configuration descriptor
                Configuration.ConfigurationDescriptor confDesc = new Configuration.ConfigurationDescriptor(500, interfaceDesc);

                // Report Descriptor
                Configuration.GenericDescriptor reportDesc = new Configuration.GenericDescriptor(Configuration.GenericDescriptor.REQUEST_Standard | Configuration.GenericDescriptor.REQUEST_Interface | Configuration.GenericDescriptor.REQUEST_IN, 0x2200, reportDescriptorPayload);

                Configuration configuration = new Configuration();

                configuration.descriptors = new Configuration.Descriptor[3];
                configuration.descriptors[0] = deviceDesc;
                configuration.descriptors[1] = confDesc;
                configuration.descriptors[2] = reportDesc;

                controllers[0].Configuration = configuration;
                

                controllers[0].Start();

                // Mouse is write on endpoint 1 only.
                mouseStream = controllers[0].CreateUsbStream(1, UsbStream.NullEndpoint);
                byte[] mouseReport = new byte[3];
                const int MOUSE_FIELD_BT = 0;
                const int MOUSE_FIELD_X = 1;
                const int MOUSE_FIELD_Y = 2;

                // Move pointer in a swirl
                const int ANGLE_STEP_SIZE = 7;
                const int MIN_CIRCLE_DIAMETER = 50;
                const int MAX_CIRCLE_DIAMETER = 300;
                const int CIRCLE_DIAMETER_STEP_SIZE = 1;

                int diameter = MIN_CIRCLE_DIAMETER;
                int diameterIncrease = CIRCLE_DIAMETER_STEP_SIZE;
                int angle = 0;
                int factor;

                screen.Clear();
                print("Connection detected...", 0, 0);
                print("The PC mouse pointer", 0, 10);
                print("will start moving soon.", 0, 20);
                screen.Flush();

                // Mouse loop Every 10ms
                while (true)
                {
                    Thread.Sleep(10);
                    if (controllers[0].Status != UsbController.PortState.Running)
                    {
                        screen.Clear();
                        print("Connection not detected...", 0, 0);
                        print("Connect EM to your PC", 0, 10);
                        print("over USB to emulate", 0, 20);
                        print("a mouse.", 0, 30);
                        screen.Flush();

                        while (controllers[0].Status != UsbController.PortState.Running) ;

                        screen.Clear();
                        print("Connection detected...", 0, 0);
                        print("The PC mouse pointer", 0, 10);
                        print("will start moving soon.", 0, 20);
                        screen.Flush();
                    }

                    // no buttons reported
                    mouseReport[MOUSE_FIELD_BT] = 0;

                    // Note X, Y are reported as change in position (relative position, not absolute)
                    factor = diameter * ANGLE_STEP_SIZE * (int)System.Math.PI / 180 / 2;
                    mouseReport[MOUSE_FIELD_X] = (byte)(-1 * factor * (int)Microsoft.SPOT.Math.Sin(angle) / 1000);
                    mouseReport[MOUSE_FIELD_Y] = (byte)(factor * (int)Microsoft.SPOT.Math.Cos(angle) / 1000);

                    angle += ANGLE_STEP_SIZE;
                    diameter += diameterIncrease;

                    if (diameter >= MAX_CIRCLE_DIAMETER ||
                        diameter <= MIN_CIRCLE_DIAMETER
                        )
                        diameterIncrease *= -1;

                    mouseStream.Write(mouseReport, 0, 3);
                }
            }
            catch (Exception e)
            {
                screen.Clear();
                print("Exception: " + e.Message, 0, 0);
                screen.Flush();

                if (mouseStream != null)
                    mouseStream.Close();
            }

            Thread.Sleep(Timeout.Infinite);
        }

        static byte[] hidClassDescriptorPayload = {
            0x01, 0x01, // bcdHID
            0x00,       // bCountryCode
            0x01,       // bNumDescriptors
            0x22,       // bDescriptorType
            0x32, 0x00  // wItemLength
        };

        static byte[] reportDescriptorPayload = {
            0x05, 0x01,     // Usage Page (Generic Desktop)
            0x09, 0x02,     // Usage (Mouse)
            0xa1, 0x01,     // Collection (Application)
            0x09, 0x01,     //   Usage (Pointer)
            0xa1, 0x00,     //   Collection (Physical)
            0x05, 0x09,     //     Usage Page (Buttons)
            0x19, 0x01,     //     Usage Minimum (01) first button
            0x29, 0x03,     //     Usage Maximun (03) third button
            0x15, 0x00,     //     Logical Minimum (0)
            0x25, 0x01,     //     Logical Maximum (1)
            0x95, 0x03,     //     Report Count (3)
            0x75, 0x01,     //     Report Size (1)
            0x81, 0x02,     //     Input (Data, Variable, Absolute)
            0x95, 0x01,     //     Report Count (1)
            0x75, 0x05,     //     Report Size (5)
            0x81, 0x01,     //     Input (Constant)    // pad for the byte
            0x05, 0x01,     //     Usage Page (Generic Desktop)
            0x09, 0x30,     //     Usage (X)
            0x09, 0x31,     //     Usage (Y)
            0x15, 0x81,     //     Logical Minimum (-127)
            0x25, 0x7f,     //     Logical Maximum (127)
            0x75, 0x08,     //     Report Size (8)
            0x95, 0x02,     //     Report Count (2)
            0x81, 0x06,     //     Input (Data, Variable, Relative)
            0xc0,           //     End Collection
            0xc0            // End Collection
        };

        // LCD Print
        static public void print(string str, int x, int y)
        {
            screen.DrawText(str, smallFont, Color.White, x, y);
        }

    }
}
