using System;
using System.Text;
using Microsoft.SPOT;

namespace GPSDecoder
{

    public class GPSDecodeUtilities
    {
        private static string decDecode = "0123456789";

        public static int GetDecValue(char ch)
        {
            int result = 0;
            foreach (char match in decDecode)
            {
                if (ch == match)
                {
                    return result;
                }
                result++;
            }
            return 0;
        }

        private static string hexDecode = "0123456789ABCDEF";

        public static int GetHexValue(char ch)
        {
            int result = 0;
            foreach (char match in hexDecode)
            {
                if (ch == match)
                {
                    return result;
                }
                result++;
            }
            return 0;
        }
    }

    public class GPSFormatErrorException : Exception
    {
        public GPSFormatErrorException(string inMessage)
            : base(inMessage)
        {
        }
    }

    public enum GPSDecoderState
    {
        waitingForStart,
        readingCommand,
        processingMessages,
        readingChecksum,
        gotResult
    }

    #region Decode machine

    public delegate void GPSUpdate ( int GPSMessageNumber );

    public class GPS : IGPS
    {
        #region Individual Message Items

        public abstract class GPSDecodeItem
        {
            protected GPS parent;

            public bool Valid;

            public int GPSMessageNumber;

            public GPSDecodeItem(GPS inParent)
            {
                parent = inParent;
            }

            public abstract void ReceiveChar(char ch);

            public abstract void StartUpdate();

            public abstract void CommitUpdate(int inGPSMessageNumber);
        }

        public class CommaSeparator : GPSDecodeItem
        {
            public CommaSeparator(GPS inParent)
                : base(inParent)
            {
                parent = inParent;
            }

            public override void ReceiveChar(char ch)
            {
                if (ch != ',')
                {
                    throw new GPSFormatErrorException("Missing separator");
                }
                Valid = true;
                parent.currentMessage.AdvanceToNextItem();
            }

            // Empty methods to allow us to create an instance of this class
            public override void StartUpdate()
            {
            }

            public override void CommitUpdate(int inGPSMessageNumber)
            {
            }
        }

        public abstract class GPSDecodeItemWithDescription : GPSDecodeItem
        {
            private string description;

            public GPSDecodeItemWithDescription(GPS inParent, string inDescription)
                : base(inParent)
            {
                description = inDescription;
            }

            public override string ToString()
            {
                return base.ToString() + " " + description;
            }
        }

        public class GPSDecodeItemFixedLengthInt : GPSDecodeItemWithDescription
        {
            private int length;
            private int currentPos;

            public GPSDecodeItemFixedLengthInt(GPS inParent, string inDescription, int inLength)
                : base(inParent, inDescription)
            {
                length = inLength;
            }

            public int Result;
            
            private int newValue;

            public override void ReceiveChar(char ch)
            {
                newValue = newValue * 10 + GPSDecodeUtilities.GetDecValue(ch);
                currentPos++;
                if (currentPos == length)
                {
                    Valid = true;
                    parent.currentMessage.AdvanceToNextItem();
                }
            }

            public override void StartUpdate()
            {
                newValue = 0;
                currentPos = 0;
            }

            public override void CommitUpdate(int inGPSMessageNumber)
            {
                Result = newValue;
                Valid = true;
                GPSMessageNumber = inGPSMessageNumber;
            }

            public override string ToString()
            {
                if (Valid)
                {
                    return base.ToString() + "  " + Result.ToString();
                }
                else
                {
                    return base.ToString() + "  ** Invalid **";
                }
            }
        }

        public class GPSDecodeItemFloat : GPSDecodeItemWithDescription
        {
            private int fractionFactor;

            private int integralPart;
            private int fractionalPart;

            public float Result;

            public GPSDecodeItemFloat(GPS inParent, string inDescription)
                : base(inParent, inDescription)
            {
            }

            public override void ReceiveChar(char ch)
            {
                if (ch == ',' || ch == '*')
                {
                    parent.currentMessage.PushBackChar();
                    parent.currentMessage.AdvanceToNextItem();
                    return;
                }

                if (ch == '.')
                {
                    fractionFactor = 1;
                    return;
                }

                if (fractionFactor != 0)
                {
                    fractionalPart = fractionalPart * 10 + GPSDecodeUtilities.GetDecValue(ch);
                    fractionFactor = fractionFactor * 10;
                    Valid = true;
                }
                else
                {
                    integralPart = integralPart * 10 + GPSDecodeUtilities.GetDecValue(ch);
                    Valid = true;
                }
            }

            public override void StartUpdate()
            {
                fractionalPart = 0;
                integralPart = 0;
                fractionFactor = 0;
                Result = 0f;
                Valid = false;
            }

            public override void CommitUpdate(int inGPSMessageNumber)
            {
                Result = integralPart;
                if (fractionalPart != 0f)
                {
                    Result = Result + ((float)fractionalPart / (float)fractionFactor);
                }
                Valid = true;
                GPSMessageNumber = inGPSMessageNumber;
            }

            public override string ToString()
            {
                if (!Valid)
                {
                    return base.ToString() + "  ** Invalid **";
                }
                else
                {
                    return base.ToString() + " " + Result.ToString();
                }
            }
        }

        public class GPSDecodeItemCoordinate : GPSDecodeItemWithDescription
        {
            private int fractionFactor;

            private int integralPart;
            private int fractionalPart;

            public int Degrees;
            public float Minutes;

            public GPSDecodeItemCoordinate(GPS inParent, string inDescription)
                : base(inParent, inDescription)
            {
            }

            public override void ReceiveChar(char ch)
            {
                if (ch == ',' || ch == '*')
                {
                    parent.currentMessage.PushBackChar();
                    parent.currentMessage.AdvanceToNextItem();
                    return;
                }

                if (ch == '.')
                {
                    fractionFactor = 1;
                    return;
                }

                if (fractionFactor != 0)
                {
                    fractionalPart = fractionalPart * 10 + GPSDecodeUtilities.GetDecValue(ch);
                    fractionFactor = fractionFactor * 10;
                    Valid = true;
                }
                else
                {
                    integralPart = integralPart * 10 + GPSDecodeUtilities.GetDecValue(ch);
                    Valid = true;
                }
            }

            public override void StartUpdate()
            {
                fractionalPart = 0;
                integralPart = 0;
                fractionFactor = 0;
                Minutes = 0f;
                Valid = false;
            }

            public override void CommitUpdate(int inGPSMessageNumber)
            {
                Degrees = integralPart / 100;
                Minutes = integralPart - (Degrees * 100);
                if (fractionalPart != 0f)
                {
                    Minutes = Minutes + ((float)fractionalPart / (float)fractionFactor);
                }
                Valid = true;
                GPSMessageNumber = inGPSMessageNumber;
            }

            public override string ToString()
            {
                if (!Valid)
                {
                    return base.ToString() + "  ** Invalid **";
                }
                else
                {
                    return base.ToString() + " Degrees:" + Degrees.ToString() + " Minutes:" + Minutes.ToString();
                }
            }
        }

        public class SingleChar : GPSDecodeItemWithDescription
        {
            public char Result;
            private char newValue;
            private bool gotNewValue;

            public SingleChar(GPS inParent, string inDescription)
                : base(inParent, inDescription)
            {
            }

            public override void ReceiveChar(char ch)
            {
                if (ch == ',' || ch == '*')
                {
                    parent.currentMessage.PushBackChar();
                }
                else
                {
                    gotNewValue = true;
                    newValue = ch;
                }
                parent.currentMessage.AdvanceToNextItem();
            }

            public override void StartUpdate()
            {
                Valid = false;
                gotNewValue = false;
            }

            public override void CommitUpdate(int inGPSMessageNumber)
            {
                if (gotNewValue)
                {
                    Result = newValue;
                    Valid = true;
                    GPSMessageNumber = inGPSMessageNumber;
                }
            }

            public override string ToString()
            {
                if (Valid)
                {
                    return base.ToString() + " " + Result.ToString();
                }
                else
                {
                    return base.ToString() + "  ** Invalid **";
                }
            }
        }

        #endregion

        #region Complete GPS Messages

        public abstract class GPSMessage
        {
            protected string Description;

            protected GPS parent;

            public char[] MessageCode;
            public string MessageName;

            public bool MatchMessageCode(char[] match, int length)
            {
                if (length != MessageCode.Length)
                {
                    return false;
                }
                for (int i = 0; i < length; i++)
                {
                    if (MessageCode[i] != match[i])
                    {
                        return false;
                    }
                }
                return true;
            }

            public GPSDecodeItem[] GPSDecodeItems;

            private GPSDecodeItem currentDecodeItem = null;

            private int currentDecodeItemNumber;

            public void AdvanceToNextItem()
            {
                currentDecodeItemNumber++;

                if (currentDecodeItemNumber == GPSDecodeItems.Length)
                {
                    parent.MessageFinished();
                    currentDecodeItem = null;
                }
                else
                {
                    currentDecodeItem = GPSDecodeItems[currentDecodeItemNumber];
                    currentDecodeItem.StartUpdate();
                }
            }

            private bool gotPushBack;

            public void PushBackChar()
            {
                gotPushBack = true;
            }

            public void ReceiveChar(char ch)
            {
                if (currentDecodeItem == null)
                {
                    return;
                }

                currentDecodeItem.ReceiveChar(ch);

                if (gotPushBack)
                {
                    if (currentDecodeItem != null)
                    {
                        currentDecodeItem.ReceiveChar(ch);
                    }
                    gotPushBack = false;
                }
            }

            public virtual void ResetMessage()
            {
                currentDecodeItemNumber = 0;
                currentDecodeItem = GPSDecodeItems[currentDecodeItemNumber];
                currentDecodeItem.StartUpdate();
            }

            public virtual void MessageComplete(int inGPSMessageNumber)
            {
                foreach (GPSDecodeItem item in GPSDecodeItems)
                {
                    item.CommitUpdate(inGPSMessageNumber);
                }
            }

            public GPSMessage(GPS inParent)
            {
                parent = inParent;
            }

            public override string ToString()
            {
                string result = "Message Name : $";

                foreach (char ch in MessageCode)
                {
                    result += ch.ToString();
                }
                result += Description;

                int itemNumber = 1;
                foreach (GPSDecodeItem item in GPSDecodeItems)
                {
                    if (item is CommaSeparator)
                    {
                        result += "\n     " + item.ToString();
                    }
                    else
                    {
                        result += "\n" + itemNumber.ToString() + " " + item.ToString();
                        itemNumber++;
                    }
                }
                return base.ToString() + result;
            }
        }

        public class GPSMessageGPGSA : GPSMessage
        {
            //    $GPGSA,A,2,01,11,20,,,,,,,,,,29.7,22.0,20.0*0E

            public GPSMessageGPGSA(GPS inParent)
                : base(inParent)
            {
                Description = "GPS DOP and active satellites";
                MessageCode = new char[] { 'G', 'P', 'G', 'S', 'A' };
                MessageName = "GPGSA";

                GPSDecodeItems = new GPSDecodeItem[] {
                    parent.manualOrAutomatic,
                    parent.comma,
                    parent.fixType,
                    parent.comma,
                    parent.prnSat1,
                    parent.comma,
                    parent.prnSat2,
                    parent.comma,
                    parent.prnSat3,
                    parent.comma,
                    parent.prnSat4,
                    parent.comma,
                    parent.prnSat5,
                    parent.comma,
                    parent.prnSat6,
                    parent.comma,
                    parent.prnSat7,
                    parent.comma,
                    parent.prnSat8,
                    parent.comma,
                    parent.prnSat9,
                    parent.comma,
                    parent.prnSat10,
                    parent.comma,
                    parent.prnSat11,
                    parent.comma,
                    parent.prnSat12,
                    parent.comma,
                    parent.pdop,
                    parent.comma,
                    parent.hdop,
                    parent.comma,
                    parent.vdop
            };
            }
        }


        /*
             $GPRMC,170319.782,A,5347.0075,N,00025.1716,W,0.000000,,050107,,*0D
             $GPRMC,hhmmss.sss,A,llll.llll,a,yyyyy.yyyy,a,x.x,x.x,ddmmyy,x.x,a,m*hh
        Field #
        1    = UTC time of fix
        2    = Data status (A=Valid position, V=navigation receiver warning)
        3    = Latitude of fix
        4    = N or S of longitude
        5    = Longitude of fix
        6    = E or W of longitude
        7    = Speed over ground in knots
        8    = Track made good in degrees True
        9    = UTC date of fix
        10   = Magnetic variation degrees (Easterly var. subtracts from true course)
        11   = E or W of magnetic variation
        12   = Mode indicator, (A=Autonomous, D=Differential, E=Estimated, N=Data not valid)
        13   = Checksum
        */

        public class GPSMessageGPRMC : GPSMessage
        {
            public GPSMessageGPRMC(GPS inParent)
                : base(inParent)
            {
                Description = "Recommended minimum specific GPS/Transit data ";
                MessageCode = new char[] { 'G', 'P', 'R', 'M', 'C' };
                MessageName = "GPRMC";

                GPSDecodeItems = new GPSDecodeItem[] {
                parent.utcHour,
                parent.utcMin,
                parent.utcSec,
                parent.utcSecFrac,
                parent.comma,
                parent.validity,
                parent.comma,
                parent.latitude,
                parent.comma,
                parent.latChar,
                parent.comma,
                parent.longitude,
                parent.comma,
                parent.longChar,
                parent.comma,
                parent.groundSpeed,
                parent.comma,
                parent.track,
                parent.comma,
                parent.utcDay,
                parent.utcMonth,
                parent.utcYear,
                parent.mageticVar,
                parent.comma,
                parent.magneticVarCh,
                parent.comma,
                parent.modeCh
            };
            }
        }

        //$GPGLL,5347.0075,N,00025.1716,W,170319.782,A*27
        public class GPSMessageGPGLL : GPSMessage
        {

            public GPSMessageGPGLL(GPS inParent)
                : base(inParent)
            {
                Description = "Geographic Position, Latitude / Longitude and time ";
                MessageCode = new char[] { 'G', 'P', 'G', 'L', 'L' };
                MessageName = "GPGLL";


                GPSDecodeItems = new GPSDecodeItem[] {
                    parent.latitude,
                    parent.comma,
                    parent.latChar,
                    parent.comma,
                    parent.longitude,
                    parent.comma,
                    parent.longChar,
                    parent.comma,
                    parent.utcHour,
                    parent.utcMin,
                    parent.utcSec,
                    parent.utcSecFrac,
                    parent.comma,
                    parent.validity
                };
            }
        }

        #endregion

        #region State Management

        private GPSDecoderState stateValue;

        public GPSDecoderState State
        {
            get
            {
                return stateValue;
            }
        }

        private GPSDecoderState state
        {
            get
            {
                return stateValue;
            }

            set
            {
                switch (value)
                {
                    case GPSDecoderState.waitingForStart:
                        break;
                    case GPSDecoderState.readingCommand:
                        checksum = 0;
                        messageCodePos = 0;
                        messageBufferPos = 0;
                        break;
                    case GPSDecoderState.processingMessages:
                        currentMessage.ResetMessage();
                        break;
                    case GPSDecoderState.readingChecksum:
                        receivedChecksum = 0;
                        checksumByteCount = 0;
                        break;
                    case GPSDecoderState.gotResult:
                        break;
                }
                stateValue = value;
            }
        }

        #endregion

        #region Data values

        private GPSDecodeItemFixedLengthInt utcHour;
        private GPSDecodeItemFixedLengthInt utcMin;
        private GPSDecodeItemFixedLengthInt utcSec;
        private GPSDecodeItemFixedLengthInt utcDay;
        private GPSDecodeItemFixedLengthInt utcMonth;
        private GPSDecodeItemFixedLengthInt utcYear;

        private GPSDecodeItemFloat utcSecFrac;
        private CommaSeparator comma;
        private SingleChar validity;
        private GPSDecodeItemCoordinate latitude;
        private SingleChar latChar;
        private GPSDecodeItemCoordinate longitude;
        private SingleChar longChar;
        private GPSDecodeItemFloat groundSpeed;
        private GPSDecodeItemFloat track;

        private GPSDecodeItemFloat mageticVar;
        private SingleChar magneticVarCh;
        private SingleChar modeCh;

        private SingleChar manualOrAutomatic;
        private SingleChar fixType;
        private GPSDecodeItemFloat prnSat1;
        private GPSDecodeItemFloat prnSat2;
        private GPSDecodeItemFloat prnSat3;
        private GPSDecodeItemFloat prnSat4;
        private GPSDecodeItemFloat prnSat5;
        private GPSDecodeItemFloat prnSat6;
        private GPSDecodeItemFloat prnSat7;
        private GPSDecodeItemFloat prnSat8;
        private GPSDecodeItemFloat prnSat9;
        private GPSDecodeItemFloat prnSat10;
        private GPSDecodeItemFloat prnSat11;
        private GPSDecodeItemFloat prnSat12;
        private GPSDecodeItemFloat pdop;
        private GPSDecodeItemFloat hdop;
        private GPSDecodeItemFloat vdop;

        private void setupMessageContent()
        {
            utcHour = new GPSDecodeItemFixedLengthInt(this, "UTC Hour Value", 2);
            utcMin = new GPSDecodeItemFixedLengthInt(this, "UTC Minute Value", 2);
            utcSec = new GPSDecodeItemFixedLengthInt(this, "UTC Second Value", 2);
            utcSecFrac = new GPSDecodeItemFloat(this, "Seconds fraction");

            comma = new CommaSeparator(this);

            validity = new SingleChar(this, "Valid-A Navigation receiver warning-V");

            latitude = new GPSDecodeItemCoordinate(this, "Latitude");
            latChar = new SingleChar(this, "N or S of latitude");
            longitude = new GPSDecodeItemCoordinate(this, "Longitude");
            longChar = new SingleChar(this, "E or W of longitude");
            groundSpeed = new GPSDecodeItemFloat(this, "Speed over ground in knots");
            track = new GPSDecodeItemFloat(this, "Track made good");


            utcDay = new GPSDecodeItemFixedLengthInt(this, "UTC Day Value", 2);
            utcMonth = new GPSDecodeItemFixedLengthInt(this, "UTC Month Value", 2);
            utcYear = new GPSDecodeItemFixedLengthInt(this, "UTC Year Value", 2);

            mageticVar = new GPSDecodeItemFloat(this, "Magnetic variation degrees");
            magneticVarCh = new SingleChar(this, "E or W of magnetic variation");
            modeCh = new SingleChar(this, "Mode indicator, (A=Autonomous, D=Differential, E=Estimated, N=Data not valid)");

            manualOrAutomatic = new SingleChar(this, "Manual-M Automatic-A");
            fixType = new SingleChar(this, "No Fix-1 2D Fix-1 3D Fix-2");
            prnSat1 = new GPSDecodeItemFloat(this,"Prn of Satellite 1");
            prnSat2 = new GPSDecodeItemFloat(this,"Prn of Satellite 2");
            prnSat3 = new GPSDecodeItemFloat(this,"Prn of Satellite 3");
            prnSat4 = new GPSDecodeItemFloat(this,"Prn of Satellite 4");
            prnSat5 = new GPSDecodeItemFloat(this,"Prn of Satellite 5");
            prnSat6 = new GPSDecodeItemFloat(this,"Prn of Satellite 6");
            prnSat7 = new GPSDecodeItemFloat(this,"Prn of Satellite 7");
            prnSat8 = new GPSDecodeItemFloat(this,"Prn of Satellite 8");
            prnSat9 = new GPSDecodeItemFloat(this,"Prn of Satellite 9");
            prnSat10 = new GPSDecodeItemFloat(this,"Prn of Satellite 10");
            prnSat11 = new GPSDecodeItemFloat(this,"Prn of Satellite 11");
            prnSat12 = new GPSDecodeItemFloat(this,"Prn of Satellite 12");
            pdop = new GPSDecodeItemFloat(this,"PDOP");
            hdop = new GPSDecodeItemFloat(this,"HDOP");
            vdop = new GPSDecodeItemFloat(this, "VDOP");

            gpsMessages = new GPSMessage[] {
                new GPSMessageGPGSA(this),
                new GPSMessageGPRMC(this),
                new GPSMessageGPGLL(this)
            };
        }

        private DateTime lastGoodReadingValue;

        public DateTime LastGoodReadingValue
        {
            get
            {
                return lastGoodReadingValue;
            }
        }

        private int GPSMessageNumber;
        private string lastMessageCodeValue;

        public string LastMessageCode
        {
            get { return lastMessageCodeValue; }
        }

        public bool GetPosition(
            ref int LatDegrees,
            ref float LatMinutes,
            ref char LatDir,
            ref int LongDegrees,
            ref float LongMinutes,
            ref char LongDir,
            ref int MessageNo)
        {
            if (validity.Result != 'A' ) 
            {
                return false;
            }
            LatDegrees = latitude.Degrees;
            LatMinutes = latitude.Minutes;
            LatDir = latChar.Result;
            LongDegrees = longitude.Degrees;
            LongMinutes = longitude.Minutes;
            LongDir = longChar.Result;
            MessageNo = GPSMessageNumber;
            return true;
        }

       private GPSUpdate updateReceivedValue;

        public GPSUpdate UpdateReceived
        {
            get { return updateReceivedValue; }
            set { updateReceivedValue = value; }
        }

        #endregion

        #region Message Processing

        private int checksum;
        private int receivedChecksum;
        private int checksumByteCount;

        private char[] messageCode = new char[20];
        private int messageCodePos;

        private char[] messageBuffer = new char[150];
        private int messageBufferPos;

        private void addToMessageBuffer(char ch)
        {
            if (messageBufferPos >= messageBuffer.Length)
            {
                return;
            }
            messageBuffer[messageBufferPos] = ch;
            messageBufferPos++;
        }

        private GPSMessage[] gpsMessages;

        private GPSMessage currentMessage;

        public void ReceiveChar(char ch)
        {
            switch (state)
            {
                case GPSDecoderState.waitingForStart:
                    if (ch == '$')
                    {
                        state = GPSDecoderState.readingCommand;
                        addToMessageBuffer(ch);
                    }
                    break;
                case GPSDecoderState.readingCommand:

                    checksum ^= ch;
                    addToMessageBuffer(ch);

                    if (ch == ',')
                    {
                        foreach (GPSMessage message in gpsMessages)
                        {
                            if (message.MatchMessageCode(messageCode, messageCodePos))
                            {
                                currentMessage = message;
                                state = GPSDecoderState.processingMessages;
                                break;
                            }
                        }
                        if (state == GPSDecoderState.readingCommand)
                        {
                            // we have not found a matching message handler - look for start
                            state = GPSDecoderState.waitingForStart;
                        }
                    }
                    else
                    {
                        if (messageCodePos < messageCode.Length)
                        {
                            messageCode[messageCodePos] = ch;
                            messageCodePos++;
                        }
                    }
                    break;
                case GPSDecoderState.processingMessages:

                    addToMessageBuffer(ch);

                    if (ch != '*')
                    {
                        checksum ^= ch;
                    }
                    try
                    {
                        currentMessage.ReceiveChar(ch);
                    }
                    catch
                    {
                        string diagnostic = "";
                        for (int i = 0; i < messageBufferPos; i++)
                        {
                            diagnostic += messageBuffer[i];
                        }
                        Debug.Print("GPS Format error : " + diagnostic);
                        state = GPSDecoderState.waitingForStart;
                    }
                    break;

                case GPSDecoderState.readingChecksum:

                    addToMessageBuffer(ch);

                    if (ch != '*')
                    {
                        receivedChecksum = (16 * receivedChecksum) + GPSDecodeUtilities.GetHexValue(ch);
                        if (checksumByteCount == 1)
                        {
                            if (receivedChecksum == checksum)
                            {
                                currentMessage.MessageComplete(++GPSMessageNumber);
                                lastGoodReadingValue = new DateTime(
                                    utcYear.Result + 2000, utcMonth.Result, utcDay.Result, 
                                    utcHour.Result, utcMin.Result, utcSec.Result);
                                lastMessageCodeValue = currentMessage.MessageName;
                                if ( UpdateReceived != null ) 
                                {
                                    UpdateReceived(GPSMessageNumber);
                                }
                            }
                            state = GPSDecoderState.waitingForStart;
                        }
                        else
                        {
                            checksumByteCount++;
                        }
                    }
                    break;
                case GPSDecoderState.gotResult:
                    break;
            }
        }

        public void MessageFinished()
        {
            state = GPSDecoderState.readingChecksum;
        }

        public void Reset()
        {
            state = GPSDecoderState.waitingForStart;
        }

        #endregion 

        public GPS()
        {
            setupMessageContent();
        }

        public override string ToString()
        {
            string result = base.ToString() + "\n";
            foreach (GPSMessage message in gpsMessages)
            {
                result += message.ToString() + "\n\n";
            }
            return result;
        }
    }

    #endregion 
}
