using System;
using System.IO;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using Microsoft.SPOT.Emulator;
using Microsoft.SPOT.Emulator.Com;

namespace Microsoft.SPOT.Emulator.MFBook
{
    public class MicroFrameworkGPSComPortComponent: EmulatorComponent
    {

        private Thread uiThread;                            // The thread managing the UI
        private GPSForm myForm;                       // The UI form associated with this component
        private ComPortToMemoryStream comStream;            // The stream through which the data will flow
        private TimerCallback rcvDataDelegate;              // Delegate to handle receiving data
        private Microsoft.SPOT.Emulator.Time.TimingServices.Timer uiRcvDataTimer;   // Timer that will fire on CLR Thread periodically to check the status of the data stream
        private TimerCallback sendDataDelegate;             // Delegate to handle sending data
        private Microsoft.SPOT.Emulator.Time.TimingServices.Timer uiSendDataTimer;  // Timer that will fire on CLR Thread when sending Data
        private byte[] byteMessage;                         // Data to send
        private object sendDataLock = new object();         // Lock for data between threads

        private delegate void CloseCOM_UIDelegate();        // Delegate called to close the UI

        /// <summary>
        /// Constructor called by the emulator
        /// </summary>
        public MicroFrameworkGPSComPortComponent()
        {
            // Create the UI
            myForm = new GPSForm(this);
        }

        /// <summary>
        /// Called by the emulator after the configuration file has been read
        /// </summary>
        public override void SetupComponent()
        {
            // Method that will be called when the UI will be closed
            myForm.Disposed += new EventHandler(myForm_Disposed);
        }

        /// <summary>
        /// Called by the emulator after all components were setup and registered
        /// </summary>
        public override void InitializeComponent()
        {
            // Find the com2 component.  A ComPortToMemoryStream in this case.
            comStream = (ComPortToMemoryStream)this.Emulator.FindComponentById("com1");

            // Setting up the periodic timer
            rcvDataDelegate = new TimerCallback(this.uiUpdateTimer_Tick);
            uiRcvDataTimer = new Microsoft.SPOT.Emulator.Time.TimingServices.Timer(this.Emulator, rcvDataDelegate, null, 0, 500);

            // Setting up the send data timer
            sendDataDelegate = new TimerCallback(this.SendData);
            uiSendDataTimer = new Microsoft.SPOT.Emulator.Time.TimingServices.Timer(this.Emulator, sendDataDelegate, null, -1, -1);

            // Setting up the thread that will handle the UI
            uiThread = new Thread(startForm);

            // Launching the UI thread.  UI should appear at this point
            uiThread.Start();
        }

        /// <summary>
        /// Method called periodically to check the incoming buffer
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void uiUpdateTimer_Tick(object stateInfo)
        {
            byte[] buffer = new byte[1024];
            int byteRead;

            // Loop until we are not receiving anymore data
            while ((byteRead = comStream.StreamOut.Read(buffer, 0, buffer.Length)) != 0)
            {
                // Convert from an array of bytes to text and display this on the UI
                myForm.AppendText(ASCIIEncoding.ASCII.GetString(buffer));
            }
        }

        /// <summary>
        /// Called by the emulator after the program exits
        /// </summary>
        public override void UninitializeComponent()
        {
            if (myForm.InvokeRequired)
            {
                // This will be called if the caller is not in the same thread as the UI thread
                myForm.Invoke(new CloseCOM_UIDelegate(myForm.Close));
            }
            else
            {
                myForm.Close();
            }
        }

        /// <summary>
        /// Method called when the form is disposed
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void myForm_Disposed(object sender, EventArgs e)
        {
            if (Emulator.IsShuttingDown == false)
                Emulator.Stop();
        }

        /// <summary>
        /// Launches the UI thread 
        /// </summary>
        private void startForm()
        {
            Application.Run(myForm);
        }


        // NOTE: Using a TimingService.Timer to switch threads is a work around in Beta3
        //       This is likely to change in RTM

        /// <summary>
        /// Gets Data from UI, and gets ready to send it to Device (UI Thread)
        /// </summary>
        /// <param name="byteMessage">The message in an array of bytes</param>
        public void SendData(byte[] byteMsg)
        {
//            Console.WriteLine(byteMsg.ToString());
            // Set lock so byteMessage isn't changed between threads
            Monitor.Enter(sendDataLock);
            byteMessage = byteMsg;

            // Cause SendData to be called once on CLR thread
            uiSendDataTimer.Change(1, -1);

        }

        /// <summary>
        /// Send data out to the device (CLR Thread)
        /// </summary>
        /// <param name="stateInfo"></param>
        private void SendData(object stateInfo)
        {
            comStream.StreamOut.Write(byteMessage, 0, byteMessage.Length);
            Monitor.Exit(sendDataLock);
        }
    }
}
